/****************************************************************************/
/* Copyright 2000 Compaq Computer Corporation.                              */
/*                                           .                              */
/* Copying or modifying this code for any purpose is permitted,             */
/* provided that this copyright notice is preserved in its entirety         */
/* in all copies or modifications.  COMPAQ COMPUTER CORPORATION             */
/* MAKES NO WARRANTIES, EXPRESSED OR IMPLIED, AS TO THE USEFULNESS          */
/* OR CORRECTNESS OF THIS CODE OR ITS FITNESS FOR ANY PARTICULAR            */
/* PURPOSE.                                                                 */
/****************************************************************************/
/*
 * btflash.c - Flash erase and program routines for Compaq Personal Server Monitor
 *
 */

#if !(defined(CONFIG_PXA) || defined(CONFIG_PXA1))
#include "sa1100.h"
#endif
#include "architecture.h"
#include "btflash.h"
#include "bootldr.h"
#include "params.h"
#include "serial.h"
#include "heap.h"
#include <errno.h>
#include <string.h>

#if defined(__linux__) || defined(__QNXNTO__)
#include <mtd-bootldr.h>
#endif

#define FLASH_TIMEOUT 20000000

#define bothbanks(w_) ((((w_)&0xFFFF) << 16)|((w_)&0xFFFF))

#ifdef CONFIG_INTEL_FLASH

static int intelFlashReset(void);
static int intelFlashBufferedProgram(unsigned long flashAddress, unsigned long *values, int nbytes);
static int intelFlashProgramBlock_1x16(unsigned long flashAddress, unsigned long *values, int nbytes);
static int intelFlashProgramWord(unsigned long flashAddress, unsigned long value);
static int intelFlashProgramWord_1x16(unsigned long flashAddress, unsigned long value);
static int intelFlashEraseChip(void);
static int intelFlashEraseSector(unsigned long sectorAddress);
static int intelFlashEraseRange(unsigned long start, unsigned long len);
static int intelFlashProtectRange(unsigned long start, unsigned long len, int protect);
static int intelFlashProtectRange_1x16(unsigned long start, unsigned long len, int protect);

static FlashAlgorithm intelFlashAlgorithm_1x16 = {
   intelFlashReset,
   intelFlashProgramWord_1x16,
   intelFlashProgramBlock_1x16,
   intelFlashEraseChip,         /* not implemented anyway */
   intelFlashEraseSector, /*  intelFlashEraseSector_1x16, */
   intelFlashEraseRange, /*  intelFlashEraseRange_1x16, */ /* will be ok if we fix erase sector */
   intelFlashProtectRange_1x16
};

static FlashAlgorithm intelFlashAlgorithm = {
       reset       : intelFlashReset,
       programWord : intelFlashProgramWord,
       programBlock: intelFlashBufferedProgram,
       eraseChip   : intelFlashEraseChip,
       eraseSector : intelFlashEraseSector,
       eraseRange  : intelFlashEraseRange,
       protectRange: intelFlashProtectRange
};

static unsigned long flashSectors_48F4400[] = {
    /* This device has 258 blocks.  It is a bottom parameter partition device,
     * so the blocks are more fine-grained for the first 18 blocks.
     *
     * The documentation on page 22 gives addresses for the map in units of
     * words (2 bytes), so the addresses here are times 2.
     */
  /* Partition 0 */
  0x00000000, 0x00010000, 0x00020000, 0x00030000,
  0x00040000, 0x00080000, 0x000c0000, 0x00100000,
  0x00140000, 0x00180000, 0x001c0000, 0x00200000,
  0x00240000, 0x00280000, 0x002c0000, 0x00300000,
  0x00340000, 0x00380000, 0x003c0000,
  /* Partition 1 */
  0x00400000, 0x00440000, 0x00480000, 0x004c0000,
  0x00500000, 0x00540000, 0x00580000, 0x005c0000,
  0x00600000, 0x00640000, 0x00680000, 0x006c0000,
  0x00700000, 0x00740000, 0x00780000, 0x007c0000,
  /* Partition 2 */
  0x00800000, 0x00840000, 0x00880000, 0x008c0000,
  0x00900000, 0x00940000, 0x00980000, 0x009c0000,
  0x00a00000, 0x00a40000, 0x00a80000, 0x00ac0000,
  0x00b00000, 0x00b40000, 0x00b80000, 0x00bc0000,
  /* Partition 3 */
  0x00c00000, 0x00c40000, 0x00c80000, 0x00cc0000,
  0x00d00000, 0x00d40000, 0x00d80000, 0x00dc0000,
  0x00e00000, 0x00e40000, 0x00e80000, 0x00ec0000,
  0x00f00000, 0x00f40000, 0x00f80000, 0x00fc0000,
  /* Partition 4 */
  0x01000000, 0x01040000, 0x01080000, 0x010c0000,
  0x01100000, 0x01140000, 0x01180000, 0x011c0000,
  0x01200000, 0x01240000, 0x01280000, 0x012c0000,
  0x01300000, 0x01340000, 0x01380000, 0x013c0000,
  /* Partition 5 */
  0x01400000, 0x01440000, 0x01480000, 0x014c0000,
  0x01500000, 0x01540000, 0x01580000, 0x015c0000,
  0x01600000, 0x01640000, 0x01680000, 0x016c0000,
  0x01700000, 0x01740000, 0x01780000, 0x017c0000,
  /* Partition 6 */
  0x01800000, 0x01840000, 0x01880000, 0x018c0000,
  0x01900000, 0x01940000, 0x01980000, 0x019c0000,
  0x01a00000, 0x01a40000, 0x01a80000, 0x01ac0000,
  0x01b00000, 0x01b40000, 0x01b80000, 0x01bc0000,
  /* Partition 7 */
  0x01c00000, 0x01c40000, 0x01c80000, 0x01cc0000,
  0x01d00000, 0x01d40000, 0x01d80000, 0x01dc0000,
  0x01e00000, 0x01e40000, 0x01e80000, 0x01ec0000,
  0x01f00000, 0x01f40000, 0x01f80000, 0x01fc0000,
  /* Partition 8 */
  0x02000000, 0x02040000, 0x02080000, 0x020c0000,
  0x02100000, 0x02140000, 0x02180000, 0x021c0000,
  0x02200000, 0x02240000, 0x02280000, 0x022c0000,
  0x02300000, 0x02340000, 0x02380000, 0x023c0000,
  /* Partition 9 */
  0x02400000, 0x02440000, 0x02480000, 0x024c0000,
  0x02500000, 0x02540000, 0x02580000, 0x025c0000,
  0x02600000, 0x02640000, 0x02680000, 0x026c0000,
  0x02700000, 0x02740000, 0x02780000, 0x027c0000,
  /* Partition 10 */
  0x02800000, 0x02840000, 0x02880000, 0x028c0000,
  0x02900000, 0x02940000, 0x02980000, 0x029c0000,
  0x02a00000, 0x02a40000, 0x02a80000, 0x02ac0000,
  0x02b00000, 0x02b40000, 0x02b80000, 0x02bc0000,
  /* Partition 11 */
  0x02c00000, 0x02c40000, 0x02c80000, 0x02cc0000,
  0x02d00000, 0x02d40000, 0x02d80000, 0x02dc0000,
  0x02e00000, 0x02e40000, 0x02e80000, 0x02ec0000,
  0x02f00000, 0x02f40000, 0x02f80000, 0x02fc0000,
  /* Partition 12 */
  0x03000000, 0x03040000, 0x03080000, 0x030c0000,
  0x03100000, 0x03140000, 0x03180000, 0x031c0000,
  0x03200000, 0x03240000, 0x03280000, 0x032c0000,
  0x03300000, 0x03340000, 0x03380000, 0x033c0000,
  /* Partition 13 */
  0x03400000, 0x03440000, 0x03480000, 0x034c0000,
  0x03500000, 0x03540000, 0x03580000, 0x035c0000,
  0x03600000, 0x03640000, 0x03680000, 0x036c0000,
  0x03700000, 0x03740000, 0x03780000, 0x037c0000,
  /* Partition 14 */
  0x03800000, 0x03840000, 0x03880000, 0x038c0000,
  0x03900000, 0x03940000, 0x03980000, 0x039c0000,
  0x03a00000, 0x03a40000, 0x03a80000, 0x03ac0000,
  0x03b00000, 0x03b40000, 0x03b80000, 0x03bc0000,
  /* Partition 15 */
  0x03c00000, 0x03c40000, 0x03c80000, 0x03cc0000,
  0x03d00000, 0x03d40000, 0x03d80000, 0x03dc0000,
  0x03e00000, 0x03e40000, 0x03e80000, 0x03ec0000,
  0x03f00000, 0x03f40000, 0x03f80000, 0x03fc0000,
  /* Partition 16 */
  0x04000000, 0x04040000, 0x04080000, 0x040c0000,
  0x04100000, 0x04140000, 0x04180000, 0x041c0000,
  0x04200000, 0x04240000, 0x04280000, 0x042c0000,
  0x04300000, 0x04340000, 0x04380000, 0x043c0000,
  /* Partition 17 */
  0x04400000, 0x04440000, 0x04480000, 0x044c0000,
  0x04500000, 0x04540000, 0x04580000, 0x045c0000,
  0x04600000, 0x04640000, 0x04680000, 0x046c0000,
  0x04700000, 0x04740000, 0x04780000, 0x047c0000,
  /* Partition 18 */
  0x04800000, 0x04840000, 0x04880000, 0x048c0000,
  0x04900000, 0x04940000, 0x04980000, 0x049c0000,
  0x04a00000, 0x04a40000, 0x04a80000, 0x04ac0000,
  0x04b00000, 0x04b40000, 0x04b80000, 0x04bc0000,
  /* Partition 19 */
  0x04c00000, 0x04c40000, 0x04c80000, 0x04cc0000,
  0x04d00000, 0x04d40000, 0x04d80000, 0x04dc0000,
  0x04e00000, 0x04e40000, 0x04e80000, 0x04ec0000,
  0x04f00000, 0x04f40000, 0x04f80000, 0x04fc0000,
  /* Partition 20 */
  0x05000000, 0x05040000, 0x05080000, 0x050c0000,
  0x05100000, 0x05140000, 0x05180000, 0x051c0000,
  0x05200000, 0x05240000, 0x05280000, 0x052c0000,
  0x05300000, 0x05340000, 0x05380000, 0x053c0000,
  /* Partition 21 */
  0x05400000, 0x05440000, 0x05480000, 0x054c0000,
  0x05500000, 0x05540000, 0x05580000, 0x055c0000,
  0x05600000, 0x05640000, 0x05680000, 0x056c0000,
  0x05700000, 0x05740000, 0x05780000, 0x057c0000,
  /* Partition 22 */
  0x05800000, 0x05840000, 0x05880000, 0x058c0000,
  0x05900000, 0x05940000, 0x05980000, 0x059c0000,
  0x05a00000, 0x05a40000, 0x05a80000, 0x05ac0000,
  0x05b00000, 0x05b40000, 0x05b80000, 0x05bc0000,
  /* Partition 23 */
  0x05c00000, 0x05c40000, 0x05c80000, 0x05cc0000,
  0x05d00000, 0x05d40000, 0x05d80000, 0x05dc0000,
  0x05e00000, 0x05e40000, 0x05e80000, 0x05ec0000,
  0x05f00000, 0x05f40000, 0x05f80000, 0x05fc0000,
  /* Partition 24 */
  0x06000000, 0x06040000, 0x06080000, 0x060c0000,
  0x06100000, 0x06140000, 0x06180000, 0x061c0000,
  0x06200000, 0x06240000, 0x06280000, 0x062c0000,
  0x06300000, 0x06340000, 0x06380000, 0x063c0000,
  /* Partition 25 */
  0x06400000, 0x06440000, 0x06480000, 0x064c0000,
  0x06500000, 0x06540000, 0x06580000, 0x065c0000,
  0x06600000, 0x06640000, 0x06680000, 0x066c0000,
  0x06700000, 0x06740000, 0x06780000, 0x067c0000,
  /* Partition 26 */
  0x06800000, 0x06840000, 0x06880000, 0x068c0000,
  0x06900000, 0x06940000, 0x06980000, 0x069c0000,
  0x06a00000, 0x06a40000, 0x06a80000, 0x06ac0000,
  0x06b00000, 0x06b40000, 0x06b80000, 0x06bc0000,
  /* Partition 27 */
  0x06c00000, 0x06c40000, 0x06c80000, 0x06cc0000,
  0x06d00000, 0x06d40000, 0x06d80000, 0x06dc0000,
  0x06e00000, 0x06e40000, 0x06e80000, 0x06ec0000,
  0x06f00000, 0x06f40000, 0x06f80000, 0x06fc0000,
  /* Partition 28 */
  0x07000000, 0x07040000, 0x07080000, 0x070c0000,
  0x07100000, 0x07140000, 0x07180000, 0x071c0000,
  0x07200000, 0x07240000, 0x07280000, 0x072c0000,
  0x07300000, 0x07340000, 0x07380000, 0x073c0000,
  /* Partition 29 */
  0x07400000, 0x07440000, 0x07480000, 0x074c0000,
  0x07500000, 0x07540000, 0x07580000, 0x075c0000,
  0x07600000, 0x07640000, 0x07680000, 0x076c0000,
  0x07700000, 0x07740000, 0x07780000, 0x077c0000,
  /* Partition 30 */
  0x07800000, 0x07840000, 0x07880000, 0x078c0000,
  0x07900000, 0x07940000, 0x07980000, 0x079c0000,
  0x07a00000, 0x07a40000, 0x07a80000, 0x07ac0000,
  0x07b00000, 0x07b40000, 0x07b80000, 0x07bc0000,
  /* Partition 31 */
  0x07c00000, 0x07c40000, 0x07c80000, 0x07cc0000,
  0x07d00000, 0x07d40000, 0x07d80000, 0x07dc0000,
  0x07e00000, 0x07e40000, 0x07e80000, 0x07ec0000,
  0x07f00000, 0x07f40000, 0x07f80000, 0x07fc0000,

  0x08000000, /* This last address is not valid, but is required because it
                 thinks there is a guard sector. */
};

static FlashDescriptor flashDescriptor_48F4400 = {
   "48F4400L0ZB0",
   bothbanks(0x89),
   bothbanks(0x8816),
   &intelFlashAlgorithm,
   sizeof(flashSectors_48F4400)/sizeof(flashSectors_48F4400[0]) - 1,
   flashSectors_48F4400,
   { "bootldr",   0x00000000, 0x00100000 - 0x00000000, LFR_BOOTLDR },
   { "kernel",    0x00100000, 0x00300000 - 0x00100000, 0 },
   { "root",      0x00300000, 0x04fc0000 - 0x00300000, 0 },
   { "home",      0x04fc0000, 0x07fc0000 - 0x04fc0000, 0 },
   { "asset",     0x07fc0000, 0x08000000 - 0x07fc0000, 0 },
};

#endif /* CONFIG_INTEL_FLASH */


static FlashDescriptor *flashDescriptors[] = {
#ifdef CONFIG_INTEL_FLASH
   &flashDescriptor_48F4400,
#endif /* CONFIG_INTEL_FLASH */
   NULL
};

static FlashDescriptor *flashDescriptors_1x16[] = {
   NULL
};

unsigned long flash_size = 0;
unsigned long flash_address_mask = -1;
int nsectors = 0;
unsigned long *flashSectors = NULL;
FlashDescriptor *flashDescriptor = NULL;


int verifiedFlashRange( unsigned long addr, unsigned long len ) {

  if ( addr + len > flash_size )
    return 0;
  else
    return 1;
}

static int checkFlashDescriptor(const char *where)
{
   if (flashDescriptor == NULL) {
      putstr("No flash descriptor: "); putstr(where); putstr("\r\n");
      return -1;
   } else {
      return 0;
   }
}

int resetFlash ()
{
   if (!checkFlashDescriptor("reset flash")) {
      return flashDescriptor->algorithm->reset();
   } else {
      return -EINVAL;
   }
}

/*
 * Programs value at flashAddress
 * Sectors must be erased before they can be programmed.
 */
int programFlashWord(unsigned long flashAddress, unsigned long value)
{
   if (!checkFlashDescriptor("program flash word")) {
      return flashDescriptor->algorithm->programWord(flashAddress, value);
   } else {
      return -EINVAL;
   }
}

/*
 * Programs block of values starting at flashAddress
 * Sectors must be erased before they can be programmed.
 */
int programFlashBlock(unsigned long flashAddress, unsigned long *values, int nbytes)
{
   if (!checkFlashDescriptor("program flash word")) {
      return flashDescriptor->algorithm->programBlock(flashAddress, values, nbytes);
   } else {
      return -EINVAL;
   }
}

int eraseFlashChip ()
{
   if (!checkFlashDescriptor("erase flash chip")) {
      return flashDescriptor->algorithm->eraseChip();
   } else {
      return -EINVAL;
   }
}

/* sectorAddress must be a valid start of sector address. sectors
   must be erased before they can be programmed! */
int eraseFlashSector (unsigned long sectorAddress)
{
   if (!checkFlashDescriptor("erase flash sector")) {
      return flashDescriptor->algorithm->eraseSector(sectorAddress);
   } else {
      return -EINVAL;
   }
}

int eraseFlashRange(unsigned long startAddress, unsigned long len)
{
   if (!checkFlashDescriptor("erase flash range")) {
      return flashDescriptor->algorithm->eraseRange(startAddress, len);
   } else {
      return -EINVAL;
   }
}

int protectFlashRange(unsigned long startAddress, unsigned long len, int protect)
{
   if (!checkFlashDescriptor("erase flash range")) {
      return flashDescriptor->algorithm->protectRange(startAddress, len, protect);
   } else {
      return -EINVAL;
   }
}

unsigned long queryFlash(unsigned long flashWordAddress)
{
   unsigned long result;
   unsigned long flashWordOffset = flashWordAddress & flash_address_mask;
   unsigned long offset = 0; /* IPCOM */

   /* put flash in query mode */
   flash_write_cmd(offset, 0x98);

   result = flash_read_val(flashWordOffset);

   /* reset flash */
   flash_write_cmd(offset, 0xff);
   return result;
}

static FlashAlgorithm *flashAlgorithm = NULL;

int updateFlashAlgorithm()
{
  int algorithm = queryFlash(0x13) & 0xFF;
  switch (algorithm) {
#ifdef CONFIG_INTEL_FLASH
  case 1:
    flashAlgorithm = &intelFlashAlgorithm;
    break;
#endif
  default:
    putLabeledWord("No flash algorithm known for CFI vendorID=", algorithm);
    algorithm = -1;
  }
  return algorithm;
}


unsigned long queryFlashID(unsigned long flashWordAddress)
{
    unsigned long result;
    unsigned long flashWordOffset = (flashWordAddress & flash_address_mask);
    int algorithm = queryFlash(0x13) & 0xFF;

    switch (algorithm) {
        case 1:
            /* Guarantee normal read mode. */
            flash_write_cmd(0x0, 0xff);
            /* put flash in read identifier codes mode */
            flash_write_cmd(0x0, 0x90);
            break;
    }

    result = flash_read_val(flashWordOffset);

    switch (algorithm) {
        case 1:
            /* Back to normal read mode. */
            flash_write_cmd(0x0, 0xff);
            break;
    }
    return result;
}

#ifdef CONFIG_INTEL_FLASH

int intelFlashReset ()
{
    /* Guarantee normal reads */
    flash_write_cmd(0x0, 0xff);
    return 0;
}

static int intelFlashReadStatus(unsigned long addr)
{
    /* address does matter */
    unsigned long offset = addr >> flash_addr_shift();
    flash_write_cmd(offset, 0x70);
    return flash_read_val(offset);
}

static int intelFlashClearStatus()
{
    /* address does not matter */
    flash_write_cmd(0, 0x50);
    return 0;
}

/* now wait for it to be programmed */
static int intelFlashWaitforStatus(unsigned long addr, long *timeoutp)
{
    long             timeout = *timeoutp;
    unsigned long    done_status;
    unsigned long    status = 0;

    done_status = flash_make_val(0x80);
    while (timeout > 0) {
        status = flash_read_val(addr);
        if ((status & done_status) == done_status) {
            break;
        }
        timeout--;
    }
    *timeoutp = timeout;
    return (status);
}

/*
 * Programs value at flashAddress
 * Sectors must be erased before they can be programmed.
 */
static int intelFlashProgramWord(unsigned long flashAddress, unsigned long value)
{
    unsigned long flashWordOffset = (flashAddress & flash_address_mask) >> 2;
    long          timeout = FLASH_TIMEOUT;
    unsigned long flashContents = nflash_read_val(flashWordOffset);
    unsigned long status = 0;

    /* see if we can program the value without erasing */
#if 0
    if ((flashContents & value) != value) {
        putstr("the flash sector needs to be erased first!\r\n");
        putLabeledWord("  flashAddress =", flashAddress);
        putLabeledWord("  flashWordOffset =", flashWordOffset);
        putLabeledWord("  flashContents=", flashContents);
        return -1;
    }
#endif

    /* Unlock the block first */
    /* Send flash the program word command */
    nflash_write_cmd(flashWordOffset, 0x40);
    nflash_write_val(flashWordOffset, value);

    /* now wait for it to be programmed */
    while(timeout > 0) {
        status = nflash_read_val(flashWordOffset);
        if((status & bothbanks(0x80)) == bothbanks(0x80)) {
            break;
        }
        timeout--;
    }
    intelFlashClearStatus();
    /* Make sure we're in normal read mode */
    nflash_write_cmd(flashWordOffset, 0xff);

    if ((timeout <= 0) || (status & bothbanks(0x7f))) {
        putstr("programFlashWord error\r\n");
        putLabeledWord("  flashAddress=", flashAddress);
        putLabeledWord("  value=", value);
        putLabeledWord("  flashContents=", flashContents);
        putLabeledWord("  status=", status);
        return(-1);
    }
    return 0;
}

#if 0
static int intelFlashProgramShort(unsigned long flashAddress, unsigned short value)
{
    long timeout = FLASH_TIMEOUT;
    unsigned short flashContents = flash_read_array(flashAddress);
    unsigned long status = 0;
    unsigned long offset = 0;  /* IPCOM Addition */

    putstr("intelFlashProgramShort: I think this is broken!\r\n");

#if 0
    /* see if we can program the value without erasing */
    if ((flashContents & value) != value) {
        putstr("the flash sector needs to be erased first!\r\n");
        putLabeledWord("  flashAddress=", flashAddress);
        putLabeledWord("  flashContents=", flashContents);
        return -1;
    }

    /* send flash the program word command */
    flash_write_cmd(offset+0x55, 0x40); /*  flashword[0x55] = bothbanks(0x40); */ /* IPCOM */
    flash_write_val(flashAddress, value); /*  flashword[flashWordOffset] = value; */
    /* now wait for it to be programmed */
    status = intelFlashWaitforStatus(offset + 0, &timeout);  /* IPCOM */
    intelFlashClearStatus();

    if ((timeout <= 0) || (status & 0x7f)) {
        putstr("programFlashWord error\r\n");
        putLabeledWord("  flashAddress=", flashAddress);
        putLabeledWord("  value=", value);
        putLabeledWord("  flashContents=", flashContents);
        putLabeledWord("  status=", status);
        return(-1);
    }
#endif
    return 0;
}
#endif

static int intelFlashProgramWord_1x16(unsigned long flashAddress, unsigned long value)
{
    putstr("intelFlashProgramWord_1x16 is not implemented\r\n");
    return -1;
#if 0
    int status;

    status = intelFlashProgramShort(flashAddress, value & 0xffff);
    if (status == 0)
        status = intelFlashProgramShort(flashAddress+2, (value>>16)&0xffff);

    return (status);
#endif
}

/*
 * Buffered Program.  The buffered program writes 64-bytes per flash
 * (128-bytes total) of data or, in other words, 32-words of data per flash.
 * We have two flashes in parallel.  It is optimal when the flashAddress is on
 * a 32-word boundary (or a 128-byte boundary).
 */
int print_addr = 0;
static int
intelFlashBufferedProgram(unsigned long flashAddress, unsigned long *values, int nbytes)
{

    unsigned long flashWordOffset = (flashAddress & flash_address_mask) >> 2;
    int nwords = nbytes >> 2;
    long timeout = FLASH_TIMEOUT;
    unsigned long status = 0;
    int i;

    if(flashWordOffset & 0x1f) {
        putstr("intelFlashBufferedProgram: flashAddress must be 128-byte aligned\r\n");
        /* It really doesn't have to, but we're doing this for speed, and it
         * is optimal to be aligned as such */
        return -1;
    }

    /* Send the Buffered Program command */
    nflash_write_cmd(flashWordOffset, 0xe8);

    /* read the Status Register */
    for(i = 0; i < 1000; i++) {
        status = nflash_read_val(flashWordOffset);
        if((status & bothbanks(0x80)) == bothbanks(0x80)) break;
    }
    if(i == 1000) {
        putstr("intelFlashBufferedProgram: Buffered Program command not accepted\n\r");
        /* Guarantee we're in normal read mode when we terminate */
        nflash_write_cmd(flashWordOffset, 0xff);
        return -1;
    }

    /* Write Word count at block start address */
    /* wordcount = length minus one */
    /* The documentation is unclear about what is written here. The text says
     * that the word count is written, but the command table (page 42) says
     * that the word count minus 1 is written.  Which is it?
     */
    nflash_write_cmd(flashWordOffset, nwords - 1);

    /* Send the Data */
    for(i = 0; i < nwords; i++) {
        nflash_write_val(flashWordOffset + i, values[i]);
/* if (0) putLabeledWord(" sr=", nflash_read_val(flashWordOffset)); */
    }

    /* Send the Confirm Command, Initiates Programming */
    nflash_write_cmd(flashWordOffset, 0xd0);

    /* now wait for it to be programmed */
    timeout = FLASH_TIMEOUT;
    while(timeout > 0) {
        status = nflash_read_val(flashWordOffset);
        if (0) putLabeledWord("  status=", status);
        if ((status & bothbanks(0x80)) == bothbanks(0x80)) break;
        timeout--;
    }
    /* Bug here: intelFlashReadStatus expects the byte offset, not the word
     * offset.  But it's not necessary anyway.  The device is already in
     * read status mode, so I'm just going to read it. */
    /* status = intelFlashReadStatus(flashWordOffset); */
    status = nflash_read_val(flashWordOffset);
    if (0) putLabeledWord("final status=", status);
    intelFlashClearStatus();
    /* Return to normal read mode */
    nflash_write_cmd(flashWordOffset & 0xFFFC0000, 0xff);

    if ((timeout <= 0) || (status & bothbanks(0x7f))) {
        putstr("programFlashBlock error\r\n");
        putLabeledWord("  flashAddress=", flashAddress);
        putLabeledWord("  status=", status);
        return -1;
    }
    return 0;
}

#define FLASH_BUFFER_SIZE  (32)

static int intelFlashProgramBlock_1x16(unsigned long flashAddress, unsigned long *long_values, int resid)
{
   unsigned long flashWordOffset = (flashAddress&flash_address_mask) >> 1;
   unsigned long blockOffset = (flashWordOffset & 0xFFFFFFF0);
   int nwords;
   long timeout = FLASH_TIMEOUT;
   unsigned long status = 0;
   int i;
   unsigned short* values = (unsigned short*)long_values;
   int  word_size_shift = flash_addr_shift();
   int  max_words;

#if 0
   putLabeledWord("intelFlashProgramFlashBlock(1x16)\r\n", flashAddress);
#endif

   nwords = resid >> word_size_shift;
   max_words = FLASH_BUFFER_SIZE >> word_size_shift;

   while (nwords) {

       int  num_to_program;

       if (nwords < max_words)
           num_to_program = nwords;
       else
           num_to_program = max_words;

       /* send the "write to buffer" command */
       flash_write_cmd(flashWordOffset, 0xe8);
       /* read the extended status register */
       do {
           status = flash_read_val(flashWordOffset);
#if 0
           putLabeledWord("  XSR=", status);
#endif
       } while ((status & 0x0080) != 0x0080);
#if 0
       putLabeledWord("final XSR=", status);
#endif

       /* write word count at block start address */
       flash_write_cmd(blockOffset, num_to_program-1);

       /* send the data */
       for (i = 0; i < num_to_program; i++) {
           flash_write_val_short(flashWordOffset+i, values[i]);
           if (0) putLabeledWord(" sr=", flash_read_val_short(blockOffset));
       }

       /* send the confirmation to program */
       flash_write_cmd(blockOffset, 0xD0);/*flashshort[blockOffset] = 0xD0; */

       /* now wait for it to be programmed */
       timeout = FLASH_TIMEOUT;
       while (timeout > 0) {
           status = flash_read_val_short(blockOffset);
           if (0) putLabeledWord("  status=", status);
           if ((status & 0x80) == 0x80)
               break;
           timeout--;
       }
       status = intelFlashReadStatus(blockOffset);
#if 0
       putLabeledWord("final status=", status);
#endif
       intelFlashClearStatus();

       if ((timeout <= 0) || (status & 0x7f)) {
           putstr("programFlashBlock error\r\n");
           putLabeledWord("  flashAddress=", flashAddress);
           putLabeledWord("  status=", status);
           return(-1);
       }

       /*
        * prepare for next iteration.
        */
       nwords -= num_to_program;
       values += num_to_program;
       flashWordOffset += num_to_program;
   }

   return 0;
}

int intelFlashEraseChip ()
{

   putstr("intelFlashEraseChip unimplemented\r\n");

   return 0;
}

/* sectorAddress must be a valid start of sector address. sectors
   must be erased before they can be programmed! */
static int intelFlashEraseSector (unsigned long sectorAddress)
{
   int i;
   long timeout = FLASH_TIMEOUT;
   unsigned long flashWordOffset = (sectorAddress & flash_address_mask);
   unsigned long flashContents;
   unsigned long status;
   unsigned long done_status;

   flashWordOffset >>= flash_addr_shift();

   for (i = 0; i < nsectors; i++) {
      if (flashSectors[i] == sectorAddress)
         break;
   }
   if (i >= nsectors) {
      putLabeledWord("eraseFlashSector: sectorAddress must be start of a sector! address=", sectorAddress);
      putLabeledWord("nsectors=", nsectors);
      return -EINVAL;
   }

#if 0
   /* Unlock the block */
   flash_write_cmd(flashWordOffset, 0x60);
   flash_write_cmd(flashWordOffset, 0xd0);
#endif
   /* send flash the erase sector command */
   flash_write_cmd(flashWordOffset, 0x20);
   flash_write_cmd(flashWordOffset, 0xd0);

   status = intelFlashWaitforStatus(flashWordOffset, &timeout);

   intelFlashClearStatus();
   flash_write_cmd(flashWordOffset, 0xff);

   done_status = flash_make_val(0x7f);

   flashContents = flash_read_array(sectorAddress);
   if ((timeout <= 0) || (status & done_status)) {
      putstr("eraseSector error\r\n");
      putLabeledWord("  sectorAddress=", sectorAddress);
      putLabeledWord("  flashContents=", flashContents);
      putLabeledWord("  status=", status);
      putLabeledWord("  timeout=", timeout);
      return(-EIO);
   }

   return 0;
}

static int intelFlashEraseRange(unsigned long startAddress, unsigned long len)
{
    unsigned long thisSectorAddress = flashSectors[0];
    unsigned long limitAddress = startAddress + len;
    unsigned long nextSectorAddress, i;

    startAddress &= flash_address_mask;
    if (limitAddress >= flashSectors[nsectors]) {
        limitAddress = flashSectors[nsectors];
    }

    for (i = 0; i < nsectors; i++) {
        nextSectorAddress = flashSectors[i + 1]; /* actually nsectors entries in this array -- last is fictitious guard sector */
        if ((thisSectorAddress <= startAddress) && (nextSectorAddress > startAddress)) {
            putLabeledWord("Erasing sector ",thisSectorAddress);
            if(eraseFlashSector(thisSectorAddress)) {
                return -1;
            }
            len -= (nextSectorAddress - startAddress);
            startAddress = nextSectorAddress;
            if (startAddress >= limitAddress)
                break;
        }
        thisSectorAddress = nextSectorAddress;
    }
    return 0;
}

static int intelFlashProtectRange(unsigned long startAddress, unsigned long len, int protect)
{

    /*
     * startAddress - the start of the address range to protect
     * len          - the len in bytes to protect
     * protect      - 1 protect blocks within the address range startAddress + len
     * protect      - 0 unprotect all blocks
     */
    putstr("intelFlashProtectRange is not implemented\n\r");
    return 0;

#if 0
    unsigned long currentSectorAddress;
    unsigned long limitAddress = startAddress + len;
    unsigned long nextSectorAddress, i;
    int result = 0;
    int status = 0;

    if (protect == 1 ) {
        putLabeledWord("startAddress :", startAddress );
        putLabeledWord("limitAddress :", limitAddress );
        for (i=0;i<nsectors;i++) {
            unsigned long timeout = FLASH_TIMEOUT;
            currentSectorAddress = flashSectors[i];
            nextSectorAddress = flashSectors[i+1]; /* actually nsectors entries in this array -- last is fictitious guard sector */
            if (((currentSectorAddress >= startAddress)
                    || (nextSectorAddress > startAddress))
                && (currentSectorAddress < limitAddress)) {
                putLabeledWord("Protecting sector ",currentSectorAddress);
                nflash_write_cmd(currentSectorAddress >> 2, 0x60); /*  flashword[currentSectorAddress >> 2] = bothbanks(0x60); */
                nflash_write_cmd(currentSectorAddress >> 2, 0x01); /*  flashword[currentSectorAddress >> 2] = bothbanks(0x01); */
                while (timeout > 0) {
                    status = nflash_read_val(currentSectorAddress >> 2);
                    if ((status & bothbanks(0x80)) == bothbanks(0x80))
                        break;
                    timeout--;
                }
                status = intelFlashReadStatus(currentSectorAddress);
                intelFlashClearStatus();
                /* put back in read mode */
                nflash_write_cmd(currentSectorAddress >> 2, 0xff); /*  flashword[currentSectorAddress >> 2] = bothbanks(0xff); */
                if (status != bothbanks(0x80)) {
                    putLabeledWord("ERROR - status=", status );
                    return -EIO;
                }
                putLabeledWord("status=", status);
                putLabeledWord("Protect=", queryFlash((currentSectorAddress>>2)+2));
            }
        }
        result = 0;
    }
    if ( protect == 0) {
        unsigned long timeout = FLASH_TIMEOUT;
        /* unprotects whole chip */

        /*
         * any address will do, but since we also use this to get
         * status after the operation, a BA is required.
         * 0 is easy to type
         */
        currentSectorAddress = 0;

        nflash_write_cmd(currentSectorAddress >> 2, 0x60); /*  flashword[currentSectorAddress >> 2] = bothbanks(0x60); */
        nflash_write_cmd(currentSectorAddress >> 2, 0xd0); /*  flashword[currentSectorAddress >> 2] = bothbanks(0xd0); */
        /* now wait for it to be complete */
        while (timeout > 0) {
            status = nflash_read_val(currentSectorAddress >> 2);
            if ((status & bothbanks(0x80)) == bothbanks(0x80))
                break;
            timeout--;
        }
        status = intelFlashReadStatus(currentSectorAddress);
        if (status != bothbanks(0x80))
            putLabeledWord("status=", status );
        result = -EIO;
        intelFlashClearStatus();
        /* put back in read mode */
        nflash_write_cmd(currentSectorAddress >> 2, 0xff); /*  flashword[currentSectorAddress >> 2] = bothbanks(0xff); */
        putLabeledWord("Protect=", queryFlash((currentSectorAddress>>2)+2));
    }
    return result;
#endif
}

/*
 * XXX this'll probably work for 2x16, too.
 * or should be able to do so easily
 */
static int intelFlashProtectRange_1x16(unsigned long startAddress, unsigned long len, int protect)
{

    /*
       startAddress - the start of the address range to protect
       len          - the len in bytes to protect
       protect      - 1 protect blocks within the address range startAddress + len
       protect      - 0 unprotect all blocks
       */

    putstr("intelFlashProtectRange_1x16 is not implemented\n\r");
    return 0;

#if 0
    unsigned long currentSectorAddress;
    unsigned long limitAddress = startAddress + len;
    unsigned long nextSectorAddress, i;
    int result = 0;
    int status = 0;
    int done_status;
#if 0 /* jca */
    unsigned short* flashshort = (unsigned short*)flashword;
#endif

    if (protect == 1 ) {
        putLabeledWord("(1x16)startAddress :", startAddress );
        putLabeledWord("(1x16)limitAddress :", limitAddress );
        for (i=0;i<nsectors;i++) {
            unsigned long timeout = FLASH_TIMEOUT;
            currentSectorAddress = flashSectors[i];
            nextSectorAddress = flashSectors[i+1]; /* actually nsectors entries in this array -- last is fictitious guard sector */
            if (((currentSectorAddress >= startAddress)
                    || (nextSectorAddress > startAddress))
                && (currentSectorAddress < limitAddress)) {
                putLabeledWord("Protecting sector ",currentSectorAddress);
#if 0
                putLabeledWord("fp(1x16), Addr=0x",
                    (unsigned long)&flashshort[currentSectorAddress >> 1]);
#endif
                flash_write_cmd(currentSectorAddress >> 1, 0x60);
                flash_write_cmd(currentSectorAddress >> 1, 0x01);
                while (timeout > 0) {
                    status = flash_read_val_short(currentSectorAddress >> 1);
                    if ((status & 0x80) == 0x80)
                        break;
                    timeout--;
                }
#if 0
                putLabeledWord("status=", status);
                putLabeledWord("timeout=", timeout);
#endif
                status = intelFlashReadStatus(currentSectorAddress);
                intelFlashClearStatus();
#if 0
                putLabeledWord("status=", status);
#endif
                /* put back in read mode */
                flash_write_val_short(currentSectorAddress >> 1, 0xff);
                if (status != 0x80) {
                    putLabeledWord("ERROR - status :", status );
                    return status;
                }
                putLabeledWord("Protect=", queryFlash((currentSectorAddress>>1)+2));
            }
        }
        result = 0;
    }
    if ( protect == 0) {
        unsigned long timeout = FLASH_TIMEOUT;
        /* unprotects whole chip */

        /*
         * any address will do, but since we also use this to get
         * status after the operation, a BA is required.
         * 0 is easy to type
         */
        /* IPCOM Any address won't do !!!! */
        currentSectorAddress = startAddress;

        flash_write_cmd(currentSectorAddress >> 1, 0x60);
        flash_write_cmd(currentSectorAddress >> 1, 0xd0);
        /* now wait for it to be complete */
        while (timeout > 0) {
            status = flash_read_val_short(currentSectorAddress >> 1);
            if ((status & 0x80) == 0x80)
                break;
            timeout--;
        }

        done_status = flash_make_val(0x80);
        status = intelFlashReadStatus(currentSectorAddress);
        if (status != done_status)
            putLabeledWord("status :", status );
        result = status;
        intelFlashClearStatus();
        /* put back in read mode */
        flash_write_val_short(currentSectorAddress >> 1, 0xff);

        putLabeledWord("Protect=", queryFlash((currentSectorAddress>>1)+2));
    }
    return result;
#endif
}

#endif /* CONFIG_INTEL_FLASH */


void btflash_print_types()
{
  int i;
  putstr("Flash types supported (2x16):\r\n");
  for (i = 0; flashDescriptors[i] != NULL; i++) {
    FlashDescriptor *fd = flashDescriptors[i];
    putstr("  "); putstr(fd->deviceName); putstr("\r\n");
  }
  putstr("Flash types supported (1x16):\r\n");
  for (i = 0; flashDescriptors_1x16[i] != NULL; i++) {
    FlashDescriptor *fd = flashDescriptors_1x16[i];
    putstr("  "); putstr(fd->deviceName); putstr("\r\n");
  }
  putstr("Current flash type is ");
  if ((flashDescriptor != NULL) &&
      (flashDescriptor->deviceName != NULL)) {
    putstr(flashDescriptor->deviceName);
    print_mem_size("\n\rFlash size: ", flash_size);
  }
  else
    putstr("UNKNOWN!\r\n");
}

void btflash_set_type(const char *devicename)
{
  int i;
  if (bt_flash_organization == BT_FLASH_ORGANIZATION_2x16) {
    for (i = 0; flashDescriptors[i] != NULL; i++) {
      FlashDescriptor *fd = flashDescriptors[i];
      if (strcmp(fd->deviceName, devicename) == 0) {
        putstr("setting flash type = "); putstr(fd->deviceName); putstr("\r\n");
        flashDescriptor = fd;

        nsectors = flashDescriptor->nsectors;
        flashSectors = flashDescriptor->sectors;
        flash_size = flashSectors[nsectors];
        flash_address_mask = flash_size-1;

        return;
      }
    }
  } else {
    for (i = 0; flashDescriptors_1x16[i] != NULL; i++) {
      FlashDescriptor *fd = flashDescriptors_1x16[i];
      if (strcmp(fd->deviceName, devicename) == 0) {
        putstr("setting flash type="); putstr(flashDescriptor->deviceName); putstr("\r\n");
        flashDescriptor = fd;

        nsectors = flashDescriptor->nsectors;
        flashSectors = flashDescriptor->sectors;
        flash_size = flashSectors[nsectors];
        flash_address_mask = flash_size-1;

        return;
      }
    }
  }
  putstr("  Unknown flash device type: "); putstr(devicename);
  putstr(" for flash organization: "); putstr(bt_flash_organization == BT_FLASH_ORGANIZATION_2x16 ? "2x16" : "1x16");
  putstr("\r\n");
}



struct BootldrFlashPartitionTable *partition_table = NULL;

void btflash_reset_partitions(void)
{
    unsigned long mach_type = MACH_TYPE;
    /* struct FlashRegion *rootP; */

    /* putLabeledWord("reset_partitions: partition_table = 0x",partition_table); */

    if (partition_table != NULL) {
        partition_table->npartitions = 0;
    }
    /* putLabeledWord("reset_partitions: flashDescriptor = 0x",flashDescriptor); */
    if (flashDescriptor != NULL) {
        btflash_define_partition(flashDescriptor->bootldr.name,
                                 flashDescriptor->bootldr.base,
                                 flashDescriptor->bootldr.size,
                                 flashDescriptor->bootldr.flags);


        btflash_define_partition(flashDescriptor->kernel.name,
                                 flashDescriptor->kernel.base,
                                 flashDescriptor->kernel.size,
                                 flashDescriptor->kernel.flags);

/* Aric */
btflash_define_partition(flashDescriptor->root.name,
                            flashDescriptor->root.base,
                            flashDescriptor->root.size,
                            flashDescriptor->root.flags);

btflash_define_partition(flashDescriptor->home.name,
                            flashDescriptor->home.base,
                            flashDescriptor->home.size,
                            flashDescriptor->home.flags);

btflash_define_partition(flashDescriptor->asset.name,
                            flashDescriptor->asset.base,
                            flashDescriptor->asset.size,
                            flashDescriptor->asset.flags);


        mach_type = param_mach_type.value;
#if 0
        if ((mach_type == MACH_TYPE_H3800) || (mach_type == MACH_TYPE_H3900)){
            rootP = btflash_get_partition(flashDescriptor->root.name);
            if (!rootP){
                putstr(" get root partition failed\r\n");
                return;
            }
            // make a 3800 specific asset partition
            btflash_define_partition("asset",
                                 rootP->base + rootP->size,
                                 256*1024,
                                 0x0);

        }
#endif
    }
}

struct FlashRegion *btflash_get_partition(const char *region_name)
{
  if (partition_table != NULL) {
    int i;
    for (i = 0; i < partition_table->npartitions; i++) {
      if (strcmp(region_name, partition_table->partition[i].name) == 0) {
        return &partition_table->partition[i];
      }
    }
    return NULL;
  } else {
    return NULL;
  }
}

unsigned long
partition_next_base(
    void)
{
    unsigned long   tnext = 0;
    unsigned long   next = 0;

    if (partition_table != NULL) {
        int i;
        for (i = 0; i < partition_table->npartitions; i++) {
            tnext = partition_table->partition[i].base +
                partition_table->partition[i].size;
            if (tnext > next) {
                next = tnext;
                /*  putLabeledWord("next: 0x", next); */
            }
        }
    }

    return (next);
}
void btflash_define_partition(const char *partition_name,
                              unsigned long base, unsigned long size,
                              enum LFR_FLAGS flags)
{
  struct FlashRegion *partition;
  int i;

#if 0
  putLabeledWord("define_partition: base 0x", base);
  putLabeledWord("define_partition: partition_table 0x", partition_table);
#endif
  if (partition_table == NULL) {
    partition_table = (BootldrFlashPartitionTable *)mmalloc(sizeof(BootldrFlashPartitionTable) + 10 * sizeof(FlashRegion));
    if (partition_table == NULL) {
      putstr("btflash_add_partition: could not allocate memory for partition table\r\n");
      return;
    }
    /* putstr("allocated partition_table\r\n"); */
    partition_table->magic = BOOTLDR_PARTITION_MAGIC;
    partition_table->npartitions = 0;
  }
  if (base == 0xffffffff)
      base = partition_next_base();

  partition = btflash_get_partition(partition_name);
  //
  for (i=partition_table->npartitions-1; i >0 ; i--){
      if (base <  partition_table->partition[i].base){
          // move it up
          partition_table->partition[i+1] = partition_table->partition[i];
          partition = &partition_table->partition[i];
      }
  }
  if (partition == NULL) {
    partition = &partition_table->partition[partition_table->npartitions];
  }
  partition_table->npartitions++;

  /* putstr("defining partition: "); putstr(partition_name); putstr("\r\n"); */
  memset(partition->name, 0, FLASH_PARTITION_NAMELEN);
  if (strlen(partition_name) < FLASH_PARTITION_NAMELEN) {
    strcpy(&partition->name[0], partition_name);
  } else {
    strncpy(&partition->name[0], partition_name, FLASH_PARTITION_NAMELEN);
    partition->name[FLASH_PARTITION_NAMELEN-1] = 0;
  }
  partition->base = base;
  partition->size = size;
  partition->flags = flags;
  if (partition->flags & LFR_EXPAND) {
      unsigned long mach_type = param_mach_type.value;
      //3800 is different, we wanna keep the asset information intact for now
      if ((mach_type == MACH_TYPE_H3800) || (mach_type == MACH_TYPE_H3900))
          partition->size = flash_size - base - 256 * 1024;
      else
          partition->size = flash_size - base;
  }
}

void btflash_delete_partition(const char *partition_name)
{
  struct FlashRegion *partition;
  int i;
  int jPart = -1;

  if (partition_table == NULL) {
      putstr("no partition table!\r\n");
      return;
  }

  partition = btflash_get_partition(partition_name);
  for (i=0; i < partition_table->npartitions; i++){
      FlashRegion *partition = &partition_table->partition[i];
      if (!strcmp(partition_name,&partition->name[0]))
          jPart = i;
  }


  if (jPart < 0) {
      putstr("no partition <");
      putstr(partition_name);
      putstr("> found\r\n");
      return;
  }

  for (i=jPart+1;i < partition_table->npartitions;i++){
      memcpy(&partition_table->partition[i-1],&partition_table->partition[i],sizeof(FlashRegion));
  }

  putstr("partition <");
  putstr(partition_name);
  putstr("> deleted\r\n");
  partition_table->npartitions--;


}


/*
 * read the CFI string and ensure that it looks like it should
 * given our flash size deduction.
 */
int
verify_flash_size(
    void)
{
    unsigned long   query_datum;
    char*           query_string = "QRY";
    int             i;
    int             query_base = 0x10;
    int             stat;

    for (i = 0; i < 3; i++) {
        stat = flash_make_val(query_string[i]);
        query_datum = queryFlash(query_base + i);
        if (stat != query_datum)
            return (0);
    }
    return (1);
}


void btflash_init()
{
    int algorithm;
    int mfrid;
    int devid;
    int protected;
    FlashDescriptor **fd;
    char* desc_type = "NONE";

    if (is_flash_16bit()) {
        fd = &flashDescriptors_1x16[0];
        desc_type = "1x16";
    }
    else {
        fd = &flashDescriptors[0];
        desc_type = "2x16";
    }

    algorithm = updateFlashAlgorithm();

    /* putstr("verify flash size...\r\n"); */
    if (!verify_flash_size()) {
        putstr("Flash Size check failed.\r\n");
        return;
    }

    mfrid = queryFlashID(0x00);
    devid = queryFlashID(0x01);
    protected = queryFlashID(0x02);

#if 0
    putLabeledWord("btflash_init: mfrid=", mfrid);
    putLabeledWord("              devid=", devid);
#endif
    if ((devid&0xFFFF) != ((devid>>16)&0xFFFF)) {
#ifndef linux
        struct bootblk_param *noeraseParam = (void *)param_noerase.value;
        putLabeledWord("\r\n\r\n@@@@@@@@ mismatched flash parts @@@@@@@@\r\n\r\n", devid);
        if (noeraseParam != NULL)
            noeraseParam->value = 1;
#endif
        flashDescriptor = fd[0];
    }
#if 0
    putLabeledWord("mfrid=", mfrid);
    putLabeledWord("deviceid=", devid);
#endif

    /* putstr("walking flash descriptors ("); */
    /* putstr(desc_type); */
    /* putstr(")\r\n"); */
    for (flashDescriptor = *fd; flashDescriptor != NULL; fd++, flashDescriptor = *fd) {
#if 0
        putstr("btflash_init: trying flash_type "); putstr(flashDescriptor->deviceName); putstr("\r\n");
        putLabeledWord("        mfrid=", flashDescriptor->manufacturerID);
        putLabeledWord("     deviceid=", flashDescriptor->deviceID);
#endif
        if (flashDescriptor->manufacturerID == mfrid
            && flashDescriptor->deviceID == devid) {
            putstr("Identified Flash: ");
            putstr(flashDescriptor->deviceName);
            putstr(" (");
            putstr(desc_type);
            putstr(")\r\n");
            nsectors           = flashDescriptor->nsectors;
            flashSectors       = flashDescriptor->sectors;
            flash_size         = flashSectors[nsectors];
            flash_address_mask = flash_size-1;

#if 0
            putLabeledWord("  flashDescriptor=", (long)flashDescriptor);
            putLabeledWord("  flashSectors=", (long)flashSectors);
            putLabeledWord("  nsectors=", nsectors);
            putLabeledWord("  flash_size=", flash_size);
            putLabeledWord("  flash_address_mask=", flash_address_mask);
#endif
            break;
        }
    }
    if (flashDescriptor == NULL) {
        putstr("Unknown flash type!\r\n");
        return;
    }
    /* Aric: This does nothing but remove a compiler warning in some cases */
    (void)intelFlashAlgorithm_1x16;
}



#ifdef CONFIG_JFFS
void btflash_jffs2_format_region(unsigned long base, unsigned long size,
                                 unsigned long marker0, unsigned long marker1, unsigned long marker2)
{
  int i;
  unsigned long limit = base + size - 12; /* leave room for 12 byte mark */
  for (i = 0; i < nsectors; i++) {
    unsigned long sectorAddress = flashSectors[i];
    if (sectorAddress >= base && sectorAddress < limit) {
      /* mark this sector for JFFS2 */
      programFlashWord(sectorAddress + 0, marker0);
      programFlashWord(sectorAddress + 4, marker1);
      programFlashWord(sectorAddress + 8, marker2);
    }
  }

}
#endif
