/*
 * Embedded reflash support for iPAQ bootloader
 *
 * Copyright (C) 2002, 2003 Phil Blundell
 *
 * Copying or modifying this code for any purpose is permitted,
 * provided that this copyright notice is preserved in its entirety
 * in all copies or modifications.
 */

#define UPPER_LEFT_ROW 50
#define UPPER_LEFT_COL 50
#define LOWER_RIGHT_ROW 220
#define LOWER_RIGHT_COL 430
#define HORIZONTAL_CENTER ((UPPER_LEFT_COL + LOWER_RIGHT_COL)/2)

#include "bootldr.h"
#include "buttons.h"
#include "commands.h"
#include "serial.h"
#include "font_8x8.c"
#include "lcd.h"
#include "hal.h"
#include "mach_type.h"
#include "pcmcia.h"
#include "vfat.h"
#include "ide.h"
#include "heap.h"
#include "params.h"
#ifdef CONFIG_MD5
#   include "md5.h"
#endif
#include "fs.h"
#include "aux_micro.h"
#include "zlib.h"
#include "zUtils.h"
#include "util.h"

#include <string.h>

#ifndef CONFIG_IDE
#error IDE support required!
#endif

#ifndef CONFIG_VFAT
#error VFAT support required!
#endif

int reflash_enabled = 0;
int direction = 0;

struct reflash_entry
{
  char *pretty_name;
  char *image_name;
  char *md5sum_name;
  char *target;
};

void
draw_text (char *s, int r, int c, int color)
{
#ifdef CONFIG_LCD
  while (*s)
    {
      unsigned const char *d = &fontdata_8x8[*s << 3];
      unsigned int x, y;

      for (y = 0; y < 8; y++)
        {
          unsigned char v = d[y];

          for (x = 0; x < 8; x++)
            {
              int val = (v & (1 << x)) ? color : 0xffff;
              /* set_pixel_value (LCD_YRES - (r + y), LCD_XRES - (c + (7 - x)), val); */
              set_pixel_value ((r + y), (c + (7 - x)), val);
            }
        }

      c += 8;
      s++;
    }
#endif
}

void
draw_rectangle (int r1, int c1, int r2, int c2, int col)
{
#ifdef CONFIG_LCD
  int r, c;
  if(r2 > r1) {
      int tmp;
      tmp = r1;
      r1 = r2;
      r2 = tmp;
  }
  if(c2 > c1) {
      int tmp;
      tmp = c1;
      c1 = c2;
      c2 = tmp;
  }
  for (r = r2; r <= r1; r++)
    {
      set_pixel_value (r, c1, col);
      set_pixel_value (r, c2, col);
    }
  for (c = c2; c <= c1; c++)
    {
      set_pixel_value (r1, c, col);
      set_pixel_value (r2, c, col);
    }
#endif
}

int
await_key(void)
{
  int r;

  set_last_buttonpress(0);

  while (r = get_last_buttonpress(), r == 0)
      ;

  return r;
}

static void
activity_callback(void)
{
#ifdef CONFIG_LCD
    static int where;

    if(where > 0) {
        lcd_fill_region(UPPER_LEFT_ROW + 52, UPPER_LEFT_COL + where,
            UPPER_LEFT_ROW + 60, UPPER_LEFT_COL + where + 8, 0xffff);
    }
    lcd_fill_region(UPPER_LEFT_ROW + 52, UPPER_LEFT_COL + where + 8,
        UPPER_LEFT_ROW + 60, UPPER_LEFT_COL + where + 16, 0x07e0);
    where += 8;
    if(UPPER_LEFT_COL + where + 16 > LOWER_RIGHT_COL - 8) {
        where -= 8;
        lcd_fill_region(UPPER_LEFT_ROW + 52, UPPER_LEFT_COL + where + 8,
            UPPER_LEFT_ROW + 60, UPPER_LEFT_COL + where + 16, 0xffff);
        where = 0;
    }
#endif
}

#define MAX_IMAGES 8
#define MAX_LEN 25

void
command_reflash (int argc, const char **argv)
{
    unsigned int detect = 0;
    char *buf = (char *)VKERNEL_BASE + 1024; /* just a temporary holding area */
    int nbytes, r;
    size_t compressed_size;
    unsigned int nr, i;
    struct reflash_entry flash[MAX_IMAGES], *fp;
    int sel;
#ifdef CONFIG_MD5
    unsigned int calc_md5[MD5_SUM_WORDS], stored_md5[MD5_SUM_WORDS];
#endif
    char *error = NULL;
    static char image_filename_store[256];
    static char target_store[64];
    static char error_message_store[256];
    int do_md5, is_gzip;
    struct bz_stream z;
    unsigned int verify_crc = 0;
    int err;
    const char *current_model = NULL;

    if (!reflash_enabled) {
        putstr ("reflashing not enabled.  Reboot with action button depressed.\r\n");
        return;
    }

do_it_again:

#ifdef CONFIG_LCD
    lcd_clear_region(UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
    draw_rectangle(
        UPPER_LEFT_ROW - 1,  UPPER_LEFT_COL - 1,
        LOWER_RIGHT_ROW + 1, LOWER_RIGHT_COL + 1,
        0xcccc
    );
    draw_rectangle(
        UPPER_LEFT_ROW - 2,  UPPER_LEFT_COL - 2,
        LOWER_RIGHT_ROW + 2, LOWER_RIGHT_COL + 2,
        0xcccc
    );
#endif

    draw_text ("Embedded Reflash Utility", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 12 * 8, 0);
    draw_text ("Scanning for images...",   UPPER_LEFT_ROW + 28, HORIZONTAL_CENTER - 11 * 8, 0x001f);

    delay_seconds(1);
    pcmcia_detect((u8*)&detect);

    set_exec_buttons_automatically(0);

    if (!detect) {
        lcd_clear_region(UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        draw_text ("No PCMCIA card found",
            UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 10 * 8, 0xf800);
        goto loop;
    }

    if(pcmcia_insert()) {
        lcd_clear_region(UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        draw_text ("No PCMCIA card found",
            UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 10 * 8, 0xf800);
        goto loop;
    }
    ide_read_ptable(NULL);

    if (r = vfat_mount_partition (0), r != 0) {
        lcd_clear_region(UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        putLabeledWord("vfat_mount_partition returned ", r);
        draw_text ("VFAT mount failed", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 8*8, 0xf800);
        goto loop;
    }

    nbytes = vfat_read_file(buf, "reflash.ctl", 0);
    if (nbytes < 0) {
        lcd_clear_region(UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        draw_text ("Can't read control file",
            UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 11*8, 0xf800);
        goto loop;
    }

#ifdef CONFIG_LCD
    lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#endif

    draw_text("Select image file to load:", UPPER_LEFT_ROW + 16, UPPER_LEFT_COL + 16, 0);

    {
        char *p = buf;
        int last = 0;

        nr = 0;

        while (!last) {
            char *pretty_name, *image_filename, *md5_filename, *target;
            char *model = NULL;

            pretty_name = p;

            while (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n' && *p != '\0') {
                p++;
            }
            if (*p == '\n' || *p == '\r') {
                p++;
                continue;
            } else if (*p == '\0') {
                break;
            }
            *p++ = 0;
            while (*p == ' ') {
                p++;
            }

            image_filename = p;

            while (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n' && *p != '\0') {
                p++;
            }
            if (*p == '\n') {
                p++;
                continue;
            } else if (*p == '\0') {
                break;
            }
            *p++ = 0;
            while (*p == ' ' || *p == '\t') {
                p++;
            }

            md5_filename = p;

            while (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n' && *p != '\0') {
                p++;
            }
            if (*p == '\n') {
                p++;
                continue;
            } else if (*p == '\0') {
                break;
            }

            *p++ = 0;
            while (*p == ' ' || *p == '\t') {
                p++;
            }

            target = p;

            while (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n' && *p != '\0') {
                p++;
            }

            if (*p == ' ' || *p == '\t') {
                *p++ = 0;
                while (*p == ' ' || *p == '\t') {
                    p++;
                }

                if (*p != '\0' && *p != '\r' && *p != '\n') {
                    model = p;

                    while (*p != ' ' && *p != '\t' && *p != '\r' && *p != '\n' && *p != '\0') {
                        p++;
                    }
                }
            }

            if (*p == 0) last = 1;

            *p++ = 0;

            if (current_model && model && strcasecmp (current_model, model)) continue;

            if (nr < MAX_IMAGES) {
                flash[nr].pretty_name = pretty_name;
                flash[nr].image_name = image_filename;
                flash[nr].md5sum_name = md5_filename;
                flash[nr].target = target;
                putLabeledWord ("image ", nr);
                putstr("name ");
                putstr(flash[nr].pretty_name);
                putstr("; image ");
                putstr(flash[nr].image_name);
                putstr("; md5sum ");
                putstr(flash[nr].md5sum_name);
                putstr("; target ");
                putstr(flash[nr].target);
                putstr("\r\n");
                nr++;
            }
        }
    }

    putLabeledWord("Total Images: ", nr);

    for(i = 0; i < nr; i++) {
        char buf[MAX_LEN];
        strncpy (buf, flash[i].pretty_name, MAX_LEN - 2);
        buf[MAX_LEN - 1] = 0;
        draw_text (buf, UPPER_LEFT_ROW + 28 + (i * 8), UPPER_LEFT_COL + 32, 0);
    }

    sel = 0;
    for(;;) {
        int k;

#ifdef CONFIG_LCD
        lcd_invert_region(UPPER_LEFT_ROW + 28 + (sel * 8), UPPER_LEFT_COL + 32,
            UPPER_LEFT_ROW + 28 + (sel * 8) + 7, UPPER_LEFT_COL + 32 + 8 * MAX_LEN);
#endif
        k = await_key();
        if (k == 5) break;

#ifdef CONFIG_LCD
        lcd_invert_region(UPPER_LEFT_ROW + 28 + (sel * 8), UPPER_LEFT_COL + 32,
            UPPER_LEFT_ROW + 28 + (sel * 8) + 7, UPPER_LEFT_COL + 32 + 8 * MAX_LEN);
#endif
        if (k == 3 && sel > 0) {
            sel--;
        } else if (k == 4 && (sel + 1) < nr) {
            sel++;
        } else if (k == 2 /* Mail button */) {
            if (direction == 0) {
                direction = 1;
                draw_text("Select image file to save:", UPPER_LEFT_ROW + 16, UPPER_LEFT_COL + 16, 0);
            } else {
                draw_text("Select image file to load:", UPPER_LEFT_ROW + 16, UPPER_LEFT_COL + 16, 0);
                direction = 0;
            }
        }
        msleep(100);
    }

#ifdef CONFIG_LCD
    lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#endif

    fp = &flash[sel];

    putstr("Selected Image is ");
    putstr(fp->pretty_name);
    putstr("\n\r");

    strcpy(image_filename_store, fp->image_name);
    strcpy(target_store, fp->target);

    if (direction != 0) {
        /* Save */
        int copyargc = 3;
        char *copyargv[3];
        /* draw_text ("Saving file...", 116 + 16, 110, 0x001f); */
        draw_text("Saving file . . .", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 8 * 8, 0);

        strcpy(image_filename_store, fp->image_name);
        copyargv[0] = "copy";
        copyargv[1] = "root";
        copyargv[2] = fp->target;

        do_command(copyargc, (const char **)&copyargv);

        draw_text("Complete", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 4 * 8, 0);
        goto loop;
    }

    draw_text("Loading file . . .", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 9 * 8, 0);
    do_md5 = strcmp(fp->md5sum_name, "-") ? 1 : 0;

#ifdef CONFIG_MD5
    if (do_md5) {
        char *lp = buf;

        /* suck in the md5sum */
        nbytes = vfat_read_file (buf, fp->md5sum_name, 0);
        if (nbytes < 0)
        {
            draw_text ("Cannot read md5sum from", UPPER_LEFT_ROW + 16, UPPER_LEFT_COL + 16, 0xf800);
            draw_text (fp->md5sum_name, UPPER_LEFT_ROW + 24, UPPER_LEFT_COL + 16, 0xf800);
            goto loop;
        }

        buf[nbytes] = 0;

        /* convert md5sum to binary form */
        for (;;)
        {
            int nibble = 0, bit = 0, offset;
            char *p;

            for (offset = 0; offset < MD5_SUM_WORDS; offset++)
                stored_md5[offset] = 0;

            for (offset = 0; offset < 32; offset++)
            {
                char c = lp[offset];
                int n;
                if (c >= '0' && c <= '9')
                    n = c - '0';
                else if (c >= 'a' && c <= 'f')
                    n = c - 'a' + 10;
                else if (c >= 'A' && c <= 'F')
                    n = c - 'A' + 10;
                else
                {
                    error = "md5sum file is corrupt";
                    goto print_error;
                }

                stored_md5[bit / 32] |= (n << ((bit % 32) + (nibble ? 0 : 4)));

                if (nibble)
                    bit += 8;

                nibble ^= 1;
            }

            lp += offset;

            if (lp[0] != ' '
                || lp[1] != ' ')
            {
                error = "md5sum file is corrupt";
                goto print_error;
            }

            lp += 2;

            p = lp;
            while (*p != '\r' && *p != '\n' && *p)
                p++;
            *(p++) = 0;

            putstr ("have md5sum for ");
            putstr (lp);
            putstr ("\r\n");

            if (strcmp (lp, image_filename_store) == 0)
                break;

            lp = p;
            while (isspace (*lp))
                lp++;
            if (*lp == 0)
            {
                error = "md5sum not found in file";
                goto print_error;
            }
        }
    }
#endif

    /* suck in the image */
    nbytes = vfat_read_file (buf, image_filename_store, 0);
    if (nbytes < 0) {
        lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        draw_text("Cannot read image from", UPPER_LEFT_ROW + 16, UPPER_LEFT_COL + 16, 0);
        draw_text(image_filename_store, UPPER_LEFT_ROW + 28, UPPER_LEFT_COL + 16, 0);
        goto loop;
    }
    putstr("Done reading.  Press Record to flash.\r\n");

#ifdef CONFIG_MD5
    if (do_md5) {
        draw_text ("Checking md5sum...", 116 + 16, 110, 0x001f);

        md5_sum(buf, nbytes, calc_md5);

        for (i = 0; i < MD5_SUM_WORDS; i++)
            if (stored_md5[i] != calc_md5[i])
            {
                error = "Image file is corrupt";
                goto print_error;
            }
    }
#endif

    is_gzip = isGZipRegion((unsigned long)buf);
    if (strcmp(target_store, "bootldr") == 0) {
        draw_text ("Checking bootldr validity", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 12*8, 0x001f);
        if (is_gzip) {
            size_t sz = nbytes;
            if (!gUnZip((unsigned long)buf, &sz, (unsigned long)buf + SZ_256K)){
                error = "Failed to gunzip bootldr image";
                goto print_error;
            }
            /* now point to the unzipped image */
            buf += SZ_256K;
            nbytes = sz;
            is_gzip = 0;
        }
        if (!isValidBootloader((unsigned long)buf, nbytes)){
            error = "Invalid bootldr into flash";
            goto print_error;
        }
    }

    if(is_gzip) {
        size_t sz = nbytes;

        draw_text("Checking data integrity", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 12*8, 0);

        if (!verifyGZipImage ((unsigned long)buf, &sz, activity_callback)) {
            error = "Error in compressed data";
            goto print_error;
        }

        /* sz now has the uncompressed size in it. */
        gzInitStream((unsigned long)buf, nbytes, &z);

        compressed_size = nbytes;
        nbytes = sz;
    }

#ifdef CONFIG_LCD
    lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#endif

    draw_text(image_filename_store, UPPER_LEFT_ROW + 16,
        HORIZONTAL_CENTER - strlen(image_filename_store)/2 * 8, 0);
    draw_text("appears to be intact.", UPPER_LEFT_ROW + 28, HORIZONTAL_CENTER - 11*8, 0);

    draw_text ("Press RECORD to flash", UPPER_LEFT_ROW + 52, HORIZONTAL_CENTER - 11*8, 0x001f);
    draw_text ("or any other key to reset", UPPER_LEFT_ROW + 64, HORIZONTAL_CENTER - 13*8, 0x001f);

    r = await_key();
    if(r != 6) {
        bootldr_reset();
    }

#ifdef CONFIG_LCD
    /* lcd_clear_region(LCD_YRES - (116 + 15), LCD_XRES - 105, 0, 0); */
    lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#endif

    {
        struct iohandle *ioh = open_iohandle(target_store);
        if(!ioh) {
            strcpy(error_message_store, "Could not open ");
            strcat(error_message_store, target_store);
            error = error_message_store;
            goto print_error;
        }

        if(ioh->ops->prewrite) {
            size_t size = ioh->ops->len(ioh);
            unsigned int i = 0;

            if(size < nbytes) {
                error = "Image does not fit in partition";
                goto print_error;
            }

            draw_text("Erasing . . .", UPPER_LEFT_ROW + 52, HORIZONTAL_CENTER - 7*8, 0x001f);

            while(i < size) {
                int l = size - i;

                if (l > 256 * 1024) l = 256 * 1024;

                if(ioh->ops->prewrite(ioh, i, l) < 0) {
                    error = "Erase failed";
                    goto print_error;
                }

                i += l;

#               ifdef CONFIG_LCD
                    if(size > 0xffff) {
                        lcd_bar(0xf800, ((i >> 16) * 100) / (size >> 16));
                    } else {
                        lcd_bar(0xf800, (i * 100) / size);
                    }
#               endif
            }
#           ifdef CONFIG_LCD
                /* Erase the bar */
                lcd_bar(0xffff, 100);
#           endif
        }

#ifdef CONFIG_LCD
        lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
        /* lcd_clear_region (LCD_YRES - (116 + 15), LCD_XRES - 105, 0, 0); */
#endif
        draw_text("Programming . . .", UPPER_LEFT_ROW + 52, HORIZONTAL_CENTER - 9*8, 0x001f);

        i = 0;
        putLabeledWord("Writing bytes: ", nbytes);
        while (i < nbytes) {
            size_t s;

            if (is_gzip) {
                unsigned char p[256];
                s = gzRead(&z, p, 256);
                if (z.err != Z_OK && z.err != Z_STREAM_END) {
                    putLabeledWord ("zerror ", z.err);
                    error = "Decompression error";
                    goto print_error;
                }
                err = ioh->ops->write(ioh, p, i, s);
            } else {
                s = nbytes - i;
                if (s > 32768)
                    s = 32768;
                err = ioh->ops->write(ioh, buf + i, i, s);
            }
            if (err < 0) {
                int len;
                memset(error_message_store, 0, sizeof(error_message_store));
                strcpy(error_message_store, "Write failure, err=");
                len = strlen(error_message_store);
                binarytohex(error_message_store+len, err, 64);
                error = error_message_store;
                goto print_error;
            }

            i += s;

#           ifdef CONFIG_LCD
                lcd_bar(0xf800, (i * 100) / nbytes);
#           endif
        }

#ifdef CONFIG_LCD
        /* Erase the bar */
        lcd_bar(0xffff, 100);
        lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#endif

        draw_text ("Verifying . . .", UPPER_LEFT_ROW + 52, HORIZONTAL_CENTER - 8*8, 0x001f);

        i = 0;
        verify_crc = 0xffffffff;
        while (i < nbytes) {
            size_t s = nbytes - i;
            static char verify_buf[1024];
            int j;

            if (s > 1024) s = 1024;

            ioh->ops->read(ioh, verify_buf, i, s);

            if (is_gzip) {
                verify_crc = crc32(verify_crc, verify_buf, s);
            } else {
                for (j = 0; j < s; j += 4) {
                    unsigned long *p1 = (unsigned long *)(buf + i + j);
                    unsigned long *p2 = (unsigned long *)(verify_buf + j);

                    if (*p1 != *p2) {
                        putLabeledWord("p1 = ", (int)p1);
                        putLabeledWord("p2 = ", (int)p2);

                        putLabeledWord("*p1 = ", *p1);
                        putLabeledWord("*p2 = ", *p2);

                        error = "Verify failure";
                        goto print_error;
                    }
                }
            }

            i += s;
#           ifdef CONFIG_LCD
            if((i & (16 * 1024 - 1)) == 0) {
                lcd_bar(0xf800, (i * 100) / nbytes);
            }
#           endif
        }
#       ifdef CONFIG_LCD
            /* Erase the bar */
            lcd_bar(0xffff, 100);
#       endif

        putstr ("Done\r\n");

        ioh->ops->close(ioh);

        if (is_gzip) {
            putLabeledWord ("nbytes ", nbytes);
            putLabeledWord ("verify_crc ", verify_crc);
            putLabeledWord ("orig_crc ", z.read_crc32);

            if (verify_crc != z.read_crc32) {
                error = "Verify failure";
                goto print_error;
            }
        }
    }

#   ifdef CONFIG_LCD
        lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#   endif

    draw_text ("Programmmed successfully", UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - 12*8, 0x001f);
    goto loop;

print_error:
#   ifdef CONFIG_LCD
        lcd_bar(0xffff, 100);
        lcd_clear_region (UPPER_LEFT_ROW, UPPER_LEFT_COL, LOWER_RIGHT_ROW, LOWER_RIGHT_COL);
#   endif
    draw_text (error, UPPER_LEFT_ROW + 16, HORIZONTAL_CENTER - strlen(error)/2 * 8, 0x001f);
    goto loop;

loop:
#   ifdef CONFIG_LCD
        lcd_bar(0xffff, 100);
#   endif
    draw_text ("Press power button to", UPPER_LEFT_ROW + 52, HORIZONTAL_CENTER - 11*8, 0x001f);
    draw_text ("restart the system.", UPPER_LEFT_ROW + 64, HORIZONTAL_CENTER - 9*8, 0x001f);
    draw_text ("Press record for", UPPER_LEFT_ROW + 76, HORIZONTAL_CENTER - 8*8, 0x001f);
    draw_text ("further reflashing", UPPER_LEFT_ROW + 88, HORIZONTAL_CENTER - 9*8, 0x001f);

    do {
        r = await_key();
        putLabeledWord ("key = ", r);
    } while((r != 1) && (r != 6));

    if(r == 6) goto do_it_again;

    if (error && (strcmp(target_store, "bootldr") == 0)) {
        /* if we got an error while programming the bootldr, it would be a bad idea to reset */
        return;
    } else {
        bootldr_reset();
    }
}

COMMAND(reflash, command_reflash, "invoke reflash utility", BB_RUN_FROM_RAM);
